<?php

namespace Cursed\Screen\Component;

use Cursed\Input\Key;
use Cursed\Screen\Renderable;
use Cursed\Screen\Screen;

class Menu implements Renderable
{
    // Double
    // const EDGE_TL = "\u{2554}";
    // const EDGE_TR = "\u{2557}";
    // const EDGE_BL = "\u{255a}";
    // const EDGE_BR = "\u{255d}";
    // const EDGE_H  = "\u{2550}";
    // const EDGE_V  = "\u{2551}";

    // Thick
    const EDGE_TL = "\u{250f}";
    const EDGE_TR = "\u{2513}";
    const EDGE_BL = "\u{2517}";
    const EDGE_BR = "\u{251b}";
    const EDGE_H  = "\u{2501}";
    const EDGE_V  = "\u{2503}";
    const EDGE_VT = "\u{2591}"; // thin, for scroll
    const EDGE_VB = "\u{2593}"; // scroll thumb
    const ICO_BULLET = "\u{2714}"; // "\u{25cf}"; // empty=25cb
    
    // const SORT_DESC = "\u{25b4}";
    // const SORT_ASC = "\u{25be}";

    const RET_DISMISS = 1;
    const RET_DISMISS_ALL = 2;

    private array $items = [];

    private int $current = 0;

    private ?string $title = null;

    public function __construct(array $items = [], ?string $title = null, int $current = 0)
    {
        array_map($this->add(...), $items);
        $this->current = $current;
        $this->title = $title;
    }

    public function setCurrent(int $current): self
    {
        $this->current = $current;
        if ($this->current < 0) {
            $this->current = 0;
        } elseif ($this->current >= count($this->items)) {
            $this->current = count($this->items) - 1;
        }
        while ($this->items[$this->current]->separator) {
            $this->current++;
        }
        return $this;
    }

    public function add(MenuItem $menuItem): self
    {
        $this->items[] = $menuItem;
        return $this;
    }

    public function render(Screen $screen, array $options = []): void
    {

        $cols = max(array_map( fn($i) => mb_strlen($i->label) + mb_strlen($i->hotkey), $this->items));
        $cols = max($cols, mb_strlen($this->title)) + 8;
        $rows = count($this->items);

        $left = $options['left'] ?? 4;
        $top = $options['top'] ?? 2;

        $screen->write("\e[0;100;37m");
        $screen->writeAt($left, $top, self::EDGE_TL.str_repeat(self::EDGE_H, $cols + 1).self::EDGE_TR);
        if ($this->title) {
            $screen->writeAt($left + 1, $top, "\e[1;7m ".$this->title." \e[22;27m");
        }
        $screen->writeAt($left, $top + 1 + $rows, self::EDGE_BL.str_repeat(self::EDGE_H, $cols + 1).self::EDGE_BR);
        for ($n = 0; $n < $rows; $n++) {
            $label = $this->items[$n]->label;
            $hotkey = $this->items[$n]->hotkey;
            $padlen = $cols-mb_strlen($label)-mb_strlen($hotkey)-5;
            if ($this->items[$n]->separator) {
                // draw separator
                if ($this->items[$n]->label) {
                    $separator = "\e[2m".str_repeat("\u{2500}", 2) . "\e[3m ".$this->items[$n]->label . " \e[23m" . str_repeat("\u{2500}", $cols - 3 - mb_strlen($this->items[$n]->label)). "\e[22m";
                } else {
                    $separator = "\e[2m".str_repeat("\u{2500}", $cols + 1)."\e[22m";
                }
                $screen->writeAt($left, $top + 1 + $n, self::EDGE_V.(($n==$this->current)?"\e[40;37m":"").$separator);
            } else {
                // draw menu item
                if (str_contains($label, "_")) {
                    $padlen += 1;
                    $label = preg_replace('<(_(.))>', "\e[4m\\2\e[24m", $label);
                }
                $pad = str_repeat(" ", $padlen);
                $checked = is_callable($this->items[$n]->isChecked) ? call_user_func($this->items[$n]->isChecked, $this->items[$n]) : $this->items[$n]->checked;
                if ($this->items[$n]->checkable) {
                    $icon = ($checked) ? " \u{2612} " : " \e[2m\u{2610}\e[22m ";
                } else {
                    if ($checked) {
                        $icon = " ".self::ICO_BULLET." ";
                    } else {
                        $icon = "   ";
                    }
                }
                $menutag = ($this->items[$n]->menu) ? " \u{25b9}": "  ";
                $screen->writeAt($left, $top + 1 + $n, self::EDGE_V.(($n==$this->current)?"\e[40;37m":"").$icon.$label.$menutag.$pad."\e[2m".$hotkey."\e[22m ");
            }
            $screen->write("\e[100;37m".self::EDGE_V);
            // $screen->writeAt($left, $top + 1 + $n, self::VERT_BAR.str_repeat(" ", $cols).self::VERT_BAR);
        }

    }

    public function handleKey(Key $key): ?int
    {
        if ($key->isLiteral) {
            foreach ($this->items as $index=>$item) {
                $label = strtolower($item->label);
                if (str_contains($label, "_".$key->key)) {
                    $this->current = $index;
                    return null;
                }
            }
        }
        switch ($key->key) {
            case Key::K_UP:
                $this->current--;
                // wrap around if we go past the top
                if ($this->current < 0) {
                    $this->current = count($this->items) - 1;
                }
                // keep tracking through separators
                while ($this->items[$this->current]?->separator) {
                    $this->current--;
                    if ($this->current < 0) {
                        $this->current = count($this->items) - 1;
                    }
                }
                break;
            case Key::K_DOWN:
                $this->current++;
                // wrap around if we go past the bottom
                if ($this->current > count($this->items) - 1) {
                    $this->current = 0;
                }
                // Keep tracking through separators if any
                while ($this->items[$this->current]->separator) {
                    $this->current++;
                    if ($this->current > count($this->items) - 1) {
                        $this->current = 0;
                    }
                }
                break;
            case Key::C_M:
                $item = $this->items[$this->current];
                if ($item->checkable) {
                    $item->checked = !$item->checked;
                }
                if (is_callable($item->callback)) {
                    call_user_func($item->callback, $item, $this);
                }
                if ($item->dismiss) {
                    return self::RET_DISMISS_ALL;
                }
                break;
        }
        return null;
    }
}
